"""
AR System Controller module

Copyright 2018 Amplifier Research
"""

import time
import visa
import re
from .errors import QueryError, ParameterError


class _AR_SC:
    """Private implementation class for common functionality between the SC1000 and SC2000.

    Use the AR_SC1000 or AR_SC2000 classes to communicate with your device.
    """
    def __init__(self):
        pass

    def initialize(self, VISAResource,
                   baudRate=9600,
                   flowControl=visa.constants.VI_ASRL_FLOW_NONE,
                   parity=visa.constants.VI_ASRL_PAR_NONE,
                   dataBits=8,
                   stopBits=visa.constants.VI_ASRL_STOP_ONE):
        """Opens a VISA session to the SC.

        If VISAResource is a serial port, it will be configured using the supplied parameters.
        flowControl, parity, and stopBits should use constants from visa.constants.
        """
        self._instr = visa.ResourceManager().open_resource(VISAResource, query_delay=0.05, write_termination='\n')

        self._instr.read_termination = '\n'

        # Configure serial if necessary
        if self._instr.interface_type == visa.constants.InterfaceType.asrl:
            self._instr.set_visa_attribute(visa.constants.VI_ATTR_ASRL_BAUD, baudRate)
            self._instr.set_visa_attribute(visa.constants.VI_ATTR_ASRL_FLOW_CNTRL, flowControl)
            self._instr.set_visa_attribute(visa.constants.VI_ATTR_ASRL_PARITY, parity)
            self._instr.set_visa_attribute(visa.constants.VI_ATTR_ASRL_DATA_BITS, dataBits)
            self._instr.set_visa_attribute(visa.constants.VI_ATTR_ASRL_STOP_BITS, stopBits)

    def reset(self):
        """Resets the SC, clears faults."""
        self._instr.write('R')

    def close(self):
        """Closes the VISA session."""
        self._instr.close()

    def set_switch(self, switch, position):
        """Set a single switch to the desired position"""
        current = self.query_switch(switch)
        if current != position:
            self._instr.write(f'SW{switch}-{position}')
            time.sleep(0.05)

    def query_switch(self, switch):
        """Queries the current position of a switch."""
        response = self._instr.query(f'SW{switch}?')
        m = re.match('SW\d+-(\d+)', response)

        if m:
            return int(m.group(1))
        else:
            raise QueryError(response)

    @staticmethod
    def _coerce(num, low, high):
        return max(low, min(high, num))


class AR_SC2000(_AR_SC):
    """SC2000"""

    def __init__(self):
        super().__init__()

    def id_query(self):
        """Returns the identity string from the SC2000."""
        return self._instr.query('*IDN?')

    @staticmethod
    def _parse_state(state, bit):
        return state & 1 << bit != 0

    def _wait_for_switching(self, timeout):
        """Reads the state continuously until switching is complete or the timeout in seconds is reached."""
        start = time.time()
        while time.time() - start < timeout:
            state = self.read_state()
            if state[0]:
                return
            else:
                time.sleep(0.5)

    def read_state(self):
        """Reads the state of the SC.

        Returns a list of boolean values representing the states shown in the manual as follows:
        [Switch operations complete, Command buffer overflow, Fault exists, Command error]
        """
        response = self._instr.query('STATE?')
        if response.startswith('STATE='):
            state = int(response.split('=')[1], 16)
            return [self._parse_state(state, 14),
                    self._parse_state(state, 13),
                    self._parse_state(state, 11),
                    self._parse_state(state, 7)]
        else:
            raise QueryError(response)

    def read_fault(self):
        """Queries the SC2000 for faults.

        0 = No fault
        1 = Fault on Interlock 1
        2 = Fault on Interlock 2
        3 = Fault on Interlock 3
        """
        response = self._instr.query('FSTA?')
        if response.startswith('FSTA='):
            return int(response.split('=')[1], 16)
        else:
            raise QueryError(response)

    def local_mode(self):
        """Sets the SC2000 to local mode."""
        self._instr.write('LOCL')

    def set_controller(self, controller):
        """Sets the active controller."""
        self._instr.write(f'CTRLR{controller}')

    def set_relay(self, bank, relay, position):
        """Sets a relay to the selected position.

        bank: integer 1 to 8
        relay: integer 1 to 16
        position: integer 0 = open, 1 = closed
        """
        bank = self._coerce(bank, 1, 8)
        relay = self._coerce(relay, 1, 16)

        if position != 0:
            position = 1

        self._instr.write(f'RB{bank}-{relay}-{position}')

    def set_open_drain(self, open_drain, active):
        """Sets the state of the specified open drain output.

        active: 0/1 or False/True.
        """
        open_drain = self._coerce(open_drain, 1, 4)
        if active in (1, True):
            active = 1
        else:
            active = 0

        self._instr.write(f'OD{open_drain}-{active}')

    def set_ttl(self, ttl, state):
        """Sets the specified TTL output low or high.

        state: 0/1 or False/True.
        """
        ttl = self._coerce(ttl, 1, 4)
        if state in (1, True):
            state = 1
        else:
            state = 0

        self._instr.write(f'TTL{ttl}-{state}')

    def set_switches(self, switches, wait_for_complete=True, timeout=10):
        """Sets multiple switches in a single command.

        switches: A list of tuples of the form (switch, position).
                e.g. [(1,2), (2,1), (3,4), (4,2)]
        wait_for_complete: If true, the state will be checked until
                the switch operations complete flag is true.
        timeout: If wait_for_complete is true, this sets the time in seconds
                to wait for switching to finish.
        """
        if all(isinstance(s, tuple)for s in switches):
            command = ''
            for sw in switches:
                command += f'SW{sw[0]}-{sw[1]};'

            self._instr.write(command)
            if wait_for_complete:
                self._wait_for_switching(timeout)
        else:
            raise TypeError('Expected a list of tuples.')

    def set_open_drains(self, open_drains, wait_for_complete=True, timeout=10):
        """Sets multiple open drains in a single command.

        open_drains: A list of tuples of the form (open_drain, active).
                e.g. [(1,0), (2,1), (3,0), (4,1)]
        wait_for_complete: If true, the state will be checked until
                the switch operations complete flag is true.
        timeout: If wait_for_complete is true, this sets the time in seconds
                to wait for switching to finish.
        """
        if all(isinstance(o, tuple)for o in open_drains):
            command = ''
            for od in open_drains:
                open_drain = self._coerce(od[0], 1, 4)
                command += f'OD{open_drain}-{od[1]};'

            self._instr.write(command)
            if wait_for_complete:
                self._wait_for_switching(timeout)
        else:
            raise TypeError('Expected a list of tuples.')

    def set_ttls(self, ttls, wait_for_complete=True, timeout=10):
        """Sets multiple TTL outputs in a single command.

        ttls: A list of tuples of the form (TTL, active).
                e.g. [(1,0), (2,1), (3,0), (4,1)]
        wait_for_complete: If true, the state will be checked until
                the switch operations complete flag is true.
        timeout: If wait_for_complete is true, this sets the time in seconds
                to wait for switching to finish.
        """
        if all(isinstance(t, tuple) for t in ttls):
            command = ''
            for ttl in ttls:
                t = self._coerce(ttl[0], 1, 4)
                command += f'TTL{t}-{ttl[1]};'

            self._instr.write(command)
            if wait_for_complete:
                self._wait_for_switching(timeout)
        else:
            raise TypeError('Expected a list of tuples.')

    def query_controller(self):
        """Queries the active controller."""
        response = self._instr.query('CTRLR?')
        m = re.match('CTRLR(\d+)', response)

        if m:
            return int(m.group(1))
        else:
            raise QueryError(response)

    def query_switch_capability(self, switch):
        """Queries the available throws for the specified switch."""
        response = self._instr.query(f'SWC{switch}?')
        m = re.match(f'SWC{switch}-(\d+)', response)

        if m:
            return int(m.group(1))
        elif response == 'NO_SWITCH':
            return 0
        else:
            raise QueryError(response)

    def query_relay(self, bank, relay):
        """Queries the state of the specified relay.

        Returns an integer, 0 = Open, 1 = Closed.
        """
        bank = self._coerce(bank, 1, 8)
        relay = self._coerce(relay, 1, 16)
        response = self._instr.query(f'RB{bank}-{relay}?')
        m = re.match(f'RB{bank}-{relay}-([01])', response)

        if m:
            return int(m.group(1))
        else:
            raise QueryError(response)

    def query_open_drain(self, open_drain):
        """Queries the current state of the specified open drain.

        Returns the state as 0 or 1.
        """
        open_drain = self._coerce(open_drain, 1, 4)
        response = self._instr.query(f'OD{open_drain}?')
        m = re.match('OD\d+-(\d)', response)

        if m:
            return int(m.group(1))
        else:
            raise QueryError(response)

    def query_ttl(self, ttl):
        """Queries the current state of the specified TTL output.

        Returns the state as 0 or 1.
        """
        ttl = self._coerce(ttl, 1, 4)
        response = self._instr.query(f'TTL{ttl}?')
        m = re.match('TTL\d+-(\d)', response)

        if m:
            return int(m.group(1))
        else:
            raise QueryError(response)

    def save_configuration(self, config):
        """Saves the present state as a pre-defined system configuration.

        config: Name of the configuration as a string, must be
                'INT1' to 'INT3', 'USER1' to 'USER8', or 'SAFE'
        """
        if re.fullmatch('(INT[1-3]|USER[1-8]|SAFE)', config.upper()):
            self._instr.write('SAVE ' + config.upper())
        else:
            raise ParameterError(config + ' is not a recognized configuration name.')

    def load_configuration(self, config, wait_for_complete=True, timeout=10):
        """Loads a pre-defined system configuration.

        config: Name of the configuration as a string, must be
                'INT1' to 'INT3', 'USER1' to 'USER8', or 'SAFE'
        wait_for_complete: If true, the state will be checked until
                the switch operations complete flag is true.
        timeout: If wait_for_complete is true, this sets the time in seconds
                to wait for switching to finish.
        """
        if re.fullmatch('(INT[1-3]|USER[1-8]|SAFE)', config.upper()):
            self._instr.write('LOAD ' + config.upper())
            if wait_for_complete:
                self._wait_for_switching(timeout)
        else:
            raise ParameterError(config + ' is not a recognized configuration name.')


class AR_SC1000(_AR_SC):
    """SC1000"""
    def __init__(self):
        super().__init__()
        self.model = 'SC1000'

    def read_fault(self):
        """Queries the SC1000 for the state of the interlock fault, returns True or False.

        Note: This function queries the GPIB status byte, so it will not work with RS-232.
        """
        if self._instr.interface_type == visa.constants.InterfaceType.gpib:
            return self._instr.read_stb() != 0
        else:
            return False

    def power(self, state):
        """Turns the SC power on or off.

        The state parameter will accept 1, 'On', or True to turn the power on.
        """
        if state in (1, True, 'ON', 'On', 'on'):
            command = 'P1'
        else:
            command = 'P0'
        self._instr.write(command)

        time.sleep(0.1)

    def set_open_collector(self, open_collector, active):
        """Sets the state of the specified open collector.

        active: 0/1 or False/True.
        """
        open_collector = self._coerce(open_collector, 1, 4)
        if active in (1, True):
            active = 1
        else:
            active = 0

        self._instr.write(f'OCT{open_collector}-{active}')

        time.sleep(0.05)

    def set_12v(self, enable):
        """Sets the state of the 12V trigger.

        enable: 0/1 or False/True.
        """
        if enable in (1, True):
            enable = 1
        else:
            enable = 0

        self._instr.write(f'+12V{enable}')

        time.sleep(0.05)

    def query_open_collector(self, open_collector):
        """Queries the current state of the specified open collector.

        Returns the state as 0 or 1.
        """
        open_collector = self._coerce(open_collector, 1, 4)
        response = self._instr.query(f'OCT{open_collector}?')
        m = re.match('OCT\d+-(\d)', response)

        if m:
            return int(m.group(1))
        else:
            raise QueryError(response)

    def query_12v(self):
        """Queries the current state of the 12V trigger.

        Returns the state as 0 or 1.
        """
        response = self._instr.query('+12V?')
        m = re.match('\+12V([01])', response)

        if m:
            return int(m.group(1))
        else:
            raise QueryError(response)
